/*   File: Hybrid.cc
     compute Zeros of holomorphic functions
     on triangulated domains

     The algorithm is based on a combination
     of validated winding number calculations
     and floating point Newton searches.

     Author: Tomas Johnson <johnson@math.uu.se>      
     Latest edit: 08/02/2008
     By: Tomas Johnson <johnson@math.uu.se>
*/

#include <fstream>
#include <cstdlib>
#include <iostream>
#include <string>
#include <stdexcept>
#include <time.h>
#include <unistd.h>
#include "list.h"

#include <cinterval.hpp>
#include <cimath.hpp>
#include "/home/nobackup/warwick/taylor/citaylor.hpp"

//#include "citaylor.hpp"

using namespace cxsc;
using namespace std;
using namespace taylor;

const interval PI(Pi());       // An enclosure of pi.
const cinterval I(interval(0), interval(1)); //An enclosure of i
const int RECNUM(10000);  //Maximal recursion depth
enum command {START_TIMING, SHOW_TIMING, STOP_TIMING};

//--------------------------------------------------------------------------
//This function checks on which side of the line between p1 and p2
//that p3 lies on.
int side(const complex &p1, const complex &p2, const complex &p3) {
  if( Re(p1) == Re(p2) )
    return sign( Re(p3) - Re(p2) );
  return sign( Im(p3)-(Re(p2)*Im(p1)-Re(p1)*Im(p2))/(Re(p2)-Re(p1))-(Im(p2)-Im(p1))/(Re(p2)-Re(p1))*Re(p3) );
}

//--------------------------------------------------------------------------
class triangle {
 public:
  //A triangle has 3 corners
  complex p1;
  complex p2;
  complex p3;
  //The winding number of the current triangle
  int wNum;

  //constructors
  triangle() {};
  triangle(const triangle &t) {p1 = t.p1; p2 = t.p2; p3 = t.p3; wNum = t.wNum;};
  triangle(complex p1in, complex p2in, complex p3in, int wNumin) 
  {p1 = p1in; p2 = p2in; p3 = p3in; wNum = wNumin;};
 
  //Calculate the size of a triangle, used for tolerance calculations
  real sizeOf() {return max((abs(p1-p2)),max((abs(p1-p3)),abs(p2-p3)));};
  //Check the largest side of the triangle, used for splitting
  int largest() {
    int i = 1; real max = abs(p1-p2); 
    if (abs(p1-p3)>max) {i=2; max=abs(p1-p3);}
    if (abs(p2-p3)>max) {i=3; max=abs(p2-p3);}
    return i;
  };
  //display the coordinates of a triangle
  void display() {cout << p1 << endl << p2 << endl << p3 << endl;};
  //Midpoint of a triangle
  complex mid() {  return (p1 + p2 + p3)/3;  };

  //Is the point p inside of the current triangle ?
  bool inside(const complex &p) {
    if( ( side(p1,p2,p3) != side(p1,p2,p) ) || 
        ( side(p3,p1,p2) != side(p3,p1,p) ) || 
        ( side(p2,p3,p1) != side(p2,p3,p) ) )
      return false;
    return true;
  };

  //Shrink a triangle towards the middle of it.
  triangle shrink(const complex &mid, const real &tol) {
    if(inside(mid)) {
      complex sp1 = mid + tol*(p1 - mid);
      complex sp2 = mid + tol*(p2 - mid);
      complex sp3 = mid + tol*(p3 - mid);
      return triangle(sp1, sp2, sp3, -1);
    }
    return triangle(p1, p2, p3, wNum);
  };
};


//--------------------------------------------------------------------------


//Timing function
//--------------------------------------------------------------------------
// 'START_TIMING' starts the clock, 'STOP_TIMING' stops the clock,
// and prints the elapsed time in hh:mm:ss. 'SHOW_TIMING' prints the
// elapsed time.
//
void clock(const command &cmd)
{
  struct tm     *ptr;          // Used for timing the program
  static time_t  time_start;   // start time

  if ( cmd == START_TIMING ) {
    time_start = time(NULL);
    ptr = localtime(&time_start);
    cout << "Started computations: " << asctime(ptr);
  }
  else if ( cmd == SHOW_TIMING || cmd == STOP_TIMING ) {
    int weeks   = 0;
    int days    = 0;
    int hours   = 0;
    int minutes = 0;
    double seconds;
    time_t time_stop;

    time_stop = time(NULL);
    ptr = localtime(&time_stop);
    if ( cmd == STOP_TIMING )
      cout << "Completed computations: " << asctime(ptr);
    seconds = difftime(time_stop, time_start);
    cout << "Elapsed run time: ";
    // Convert seconds to vv:dd:hh:mm:ss
    while ( seconds >= 60 ) { seconds -= 60; minutes++; }
    while ( minutes >= 60 ) { minutes -= 60;   hours++; }
    while (   hours >= 24 ) {   hours -= 24;    days++; }
    while (    days >=  7 ) {    days -=  7;   weeks++; }
    if ( weeks > 0)
      cout << weeks << " week" << (weeks != 1 ? "s, " : ", ");
    if ( days > 0)
      cout << days << " day" << (days != 1 ? "s, " : ", ");
    if ( hours > 0)
      cout << hours << " hour" << (hours != 1 ? "s, " : ", ");
    if ( minutes > 0)
      cout << minutes << " minute" << (minutes != 1 ? "s, " : ", ");
    cout << seconds << " second" << (seconds != 1 ? "s." : ".") << endl;
  }
}

//--------------------------------------------------------------------------
//Parametrise a line
citaylor zoft(const citaylor &t, const complex &p1, const complex &p2) {
  return p1 + t*(p2-p1);
}

////////////////////////////////////////////////////////////////
//Auxiliary power function, to be used for z^n, n\in N
//The reason is that there is no special method in citaylor
//to compute the power function for this case. I.e. z^n is calculated
//as exp(n*log(z)) which can cause (non-mathematical) branch cut 
//problems when the derivatives of z^n are calculated.
//This implementation is (perhaps) non-optimal, but gives a method to 
//enclose the derivatives of z^n without causing run-time errors.
//In case that the real part of z does not cut Re(z)=0
//AND the real part of the box is large enough (completely heuristic !)
//it computes the power function using the one included in citaylor.
//Otherwise the function computes z^n recursively as:
//           z^n = z^(mod(n,2))*sqr(z^(floor(n/2)));
///////////////////////////////////////////////////////////////////
citaylor aPow(const citaylor &z, const int &n) {
  if ( !(0.0 <= get_j_coef(z,0)) && (diam(Re(get_j_coef(z,0)))>0.00001) ) {
    citaylor res=power(z,n);
    return res;
  }
  
  if (n==1)
    return z;
  if (n==2)
    return sqr(z);
  int k=n/2;
  if (n==2*k) 
    return sqr(aPow(z,k));
  else
    return z*sqr(aPow(z,k));
  
}

////////////////////////////////////////////////////////////////////////////
//The function, whose zeros are calculated
citaylor fcnF(const citaylor &z) {
  //    return aPow(z,11)-1;
  
return (aPow(z-interval(0.0067)*I,37)-1/sqrt(interval(2.0)))*(aPow(z,200)-interval(1.1));
  //return (aPow(z,50)+aPow(z,12)-5*sin(20*z)*cos(12*z)-1)*sin(sqr(z)/((sqr(z)-sqr(I+1))*(sqr(z)-sqr(1-I))));
  
  //Herlocker, Ely:
  //return power(z,3) - sqr(z) - z - 2; //(1)

  //Petkovic
  //return interval(70)*sqr(sqr(z)-2*z+3)*(z-1-interval(99)/interval(70)*I); //(2)

  //Dellnitz, Schtze, Zheng
  //return aPow(z,50)+aPow(z,12)-5*sin(20*z)*cos(12*z)-1;  //(3)
  //return sqr(z)+interval(-19435)/interval(pow(10,6))*z+interval(100041)/interval(100)*
  // exp(-interval(5)/interval(1000)*z)+interval(522463); //(4)


  //As in Ko et al
  //return sqr(sqr(z)+z+1)*aPow(z-1,4)*aPow(z*sqr(z)+sqr(z)+z+1,3)*(z-2)*aPow(z-4,4); //(18s)
  //return interval(1.0)-sqr(aPow(z,16)-1); //(11s)
  
  //Test Resolving a cluster
  //return (sqr(z)-1)*(z-I);
}

////////////////////////////////////////////////////////////
//Compute the winding number, using 
//\sum_{discretisation}Arg(f(c_{i+1})/f(c_i))
interval simInt(const interval &t, const complex &lC, const complex &rC, bool &ok, const int &order) 
{
  cinterval rightleftF; 
  citaylor tA(order-1, t);
  citaylor tr(order-1, Sup(t));
  citaylor tl(order-1, Inf(t));
  citaylor z=zoft(tA, lC, rC);  //Entire line
  citaylor rightz=zoft(tr, lC, rC);  //Right end point of the line
  citaylor leftz=zoft(tl, lC, rC);  //Left end point of the line
  
  cinterval fVal=get_j_coef(fcnF(z), order-1);
  //Check if the winding of the line segment is less than pi
  if( 0.0<=Re(fVal) && 0.0<=Im(fVal) ){
    ok=false;
    return t;
  }
  //Compute the increment of the argument
  rightleftF=get_j_coef(fcnF(rightz), order-1)/get_j_coef(fcnF(leftz),order-1);

  //Forces initial, minimum discretisation
  //Arg, is the principal value argument function
  if (Inf(Re(rightleftF))<=0.0 && 0<=Im(rightleftF) ) {
    ok=false;
    return t;
  }
  
  //Return the increment of the argument, normalised.
  return Arg(rightleftF)/(2*PI);

}//--------------------------------------------------------------------------


//----------------------------------------------------------------------------
//Compute the contribution to the winding number along one side of the triangle
//This function controls the recursion.
//fail = 0, to many iterations
//fail = 1, OK
interval quadSide(const interval &t, const complex &p1, const complex &p2, const real &tol, 
		  bool &fail, const int &rec, const int &order) 
{
  if (fail == false || (rec > RECNUM) ) {
    fail = false;
    return t;
  } 
    
  bool ok(true);
  //Calculate the winding
  interval res = simInt(t, p1, p2, ok, order);
  
  //If the calculation fails, split the line in two and redo
  if( !ok || diam(res)>tol ) {
    res = quadSide(interval(Inf(t),mid(t)), p1, p2, tol/2, fail, rec+1,order)
        + quadSide(interval(mid(t),Sup(t)), p1, p2, tol/2, fail, rec+1,order); 
  }
  return res;
}

//--------------------------------------------------------------------------
//Compute the winding number of a triangle.
interval quad(const triangle &Dom, bool &fail, const int &order) 
{
  interval res(0);
  real tol(0.32);
  res = quadSide(interval(0,1), Dom.p1, Dom.p2, tol, fail, 0, order) //First  side
    + quadSide(interval(0,1), Dom.p2, Dom.p3, tol, fail, 0, order)   //Second side
    + quadSide(interval(0,1), Dom.p3, Dom.p1, tol, fail, 0, order);  //Third  side
  return res;
}

//--------------------------------------------------------------------------
// Bisects along the largest side.
//
void splitAndStore(triangle &z, List<triangle> &workList)
{
  triangle newTriangle1, newTriangle2;
  int oldwNum=z.wNum;
  int i=z.largest(); //Choose the longest side
  if (i==1) {
    cinterval tempP(z.p1+z.p2);
    tempP/=2.0;
    complex pnew(Inf(Re(tempP)),Inf(Im(tempP)));
    real dist=abs(pnew-z.p3);
    if (dist<abs(complex(Inf(Re(tempP)),Sup(Im(tempP)))-z.p3)) {
      pnew=complex(Inf(Re(tempP)),Sup(Im(tempP)));
      dist=abs(complex(Inf(Re(tempP)),Sup(Im(tempP)))-z.p3);
    }
    if (dist<abs(complex(Sup(Re(tempP)),Inf(Im(tempP)))-z.p3)) {
      pnew=complex(Sup(Re(tempP)),Inf(Im(tempP)));
      dist=abs(complex(Sup(Re(tempP)),Inf(Im(tempP)))-z.p3);
    }
    if (dist<abs(complex(Sup(Re(tempP)),Sup(Im(tempP)))-z.p3)) {
      pnew=complex(Sup(Re(tempP)),Sup(Im(tempP)));
    }

    newTriangle1=triangle(z.p1, pnew, z.p3, -1);
    newTriangle2=triangle(pnew, z.p2, z.p3, -1);
  }
  if (i==2) {
    cinterval tempP(z.p1+z.p3);
    tempP/=2.0;
    complex pnew(Inf(Re(tempP)),Inf(Im(tempP)));
    real dist=abs(pnew-z.p2);
    if (dist<abs(complex(Inf(Re(tempP)),Sup(Im(tempP)))-z.p2)) {
      pnew=complex(Inf(Re(tempP)),Sup(Im(tempP)));
      dist=abs(complex(Inf(Re(tempP)),Sup(Im(tempP)))-z.p2);
    }
    if (dist<abs(complex(Sup(Re(tempP)),Inf(Im(tempP)))-z.p2)) {
      pnew=complex(Sup(Re(tempP)),Inf(Im(tempP)));
      dist=abs(complex(Sup(Re(tempP)),Inf(Im(tempP)))-z.p2);
    }
    if (dist<abs(complex(Sup(Re(tempP)),Sup(Im(tempP)))-z.p2)) {
      pnew=complex(Sup(Re(tempP)),Sup(Im(tempP)));
    }
    
    newTriangle1=triangle(z.p1, z.p2 , pnew, -1);
    newTriangle2=triangle(pnew, z.p2, z.p3, -1);
  }
  
  if (i==3) {
    cinterval tempP(z.p2+z.p3);
    tempP/=2.0;
    complex pnew(Inf(Re(tempP)),Inf(Im(tempP)));
    real dist=abs(pnew-z.p1);
    if (dist<abs(complex(Inf(Re(tempP)),Sup(Im(tempP)))-z.p1)) {
      pnew=complex(Inf(Re(tempP)),Sup(Im(tempP)));
      dist=abs(complex(Inf(Re(tempP)),Sup(Im(tempP)))-z.p1);
    }
    if (dist<abs(complex(Sup(Re(tempP)),Inf(Im(tempP)))-z.p1)) {
      pnew=complex(Sup(Re(tempP)),Inf(Im(tempP)));
      dist=abs(complex(Sup(Re(tempP)),Inf(Im(tempP)))-z.p1);
    }
    if (dist<abs(complex(Sup(Re(tempP)),Sup(Im(tempP)))-z.p1)) {
      pnew=complex(Sup(Re(tempP)),Sup(Im(tempP)));
    }
    
    newTriangle1=triangle(z.p1, pnew, z.p3, -1);
    newTriangle2=triangle(z.p1, z.p2, pnew, -1);
  }

  bool fail(1);
  interval arg(0);
  //Calculate the winding number for one of the new triangles
  arg=quad(newTriangle1, fail, 1);

  if (!fail && (ceil(_double(Inf(arg)))==floor(_double(Sup(arg))) ) ) {
    newTriangle1.wNum=abs((int)ceil(_double(Inf(arg))));
    if ( newTriangle1.wNum!=0 )  {
      workList+=newTriangle1;
      if ( newTriangle1.wNum<oldwNum ) {
	newTriangle2.wNum=oldwNum-newTriangle1.wNum;
	workList+=newTriangle2;
      }
    }
    else {
      newTriangle2.wNum=oldwNum;
      workList+=newTriangle2;
    }
  }
  else {
    newTriangle1.wNum=-1;
    newTriangle2.wNum=-1;
    workList+=newTriangle1;
    workList+=newTriangle2;
  }

}
//--------------------------------------------------------------------------
//Search for zeros using Newtons method for 'f^(NZ-1)', where 'NZ'
//is the number of zeros in the triangle 'Dom'

bool searchZero(triangle &Dom, int NZ, const int &ntolp) {
  //Don't search for zeros of too high order
  if (NZ>100)   //Choose this constant wisely !
    return false;
  try {
    real eps(pow(10,-ntolp));  //Choosen resolution
    citaylor z(NZ,Dom.mid());   
    citaylor f=fcnF(z);
    
    complex zz;
    if ( 0 <= get_j_coef(f,NZ) )
      return false;
    else
      zz = Dom.mid() - mid(get_j_derive(f,NZ-1))/mid(get_j_derive(f,NZ));
    //Check that the updated approximation is inside of the domain
    if ( !(Dom.inside(zz)) ) 
      return false;
    
    int    i = 0;
    int nits = max(5,ntolp);

    //Redo 'nits' Newton iterations
    while( i < nits && Dom.inside(zz) ) {
      citaylor zt(NZ, zz);
      f = fcnF(zt);
      if ( (0 <=get_j_coef(f,NZ)) ) 
	i = nits;
      else {
	complex zztemp = zz - mid(get_j_derive(f,NZ-1))/mid(get_j_derive(f,NZ));
	if( Dom.inside(zztemp) ) {
	  zz = zztemp;
	   i = i + 1;
	}
	else
	  i = nits;
      }
    }

    bool fail(true);
    triangle Dom2;
    interval arg;
   
    //Shrink the domain
    Dom2 = Dom.shrink(zz, eps);
    //Validate that the shrunken domain

    //still encloses the zero.
    arg  = quad(Dom2, fail, 1);
    
    if( fail != true)
      return false;
    
    if (    (ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) 
	    && (abs((int)floor(_double(Sup(arg)))) == NZ) ) {
      //Return the shrunken domain
      Dom = Dom2;
      Dom.wNum=NZ;
      return true;
    }
  }
  catch(...){}
  return false;
}

//--------------------------------------------------------------------------
//Try to distinguish between a cluster and a multiple zero
bool MultipleRoot(const triangle &Dom, const int &NZ) {
  bool fail(true);
  //Check if f^(NZ-1) has a zero in the current triangle
  interval arg = quad(Dom, fail, NZ); 
  
  if( fail == false )
    return true;
  if ( ceil(_double(Inf(arg))) == floor(_double(Sup(arg))) ) {
    int resNZ = abs((int)ceil(_double(Inf(arg))));
    if( resNZ != 0 )
      return true;
    //If no root of order (NZ) return false
    else
      return false;
  }
  else
    return true;
}

//--------------------------------------------------------------------------
//Read the triangulation from disk
//The triangulation should be on the form
//x1 y1 x2 y2 x3 y3

void initiate(List<triangle> &workList, string inFileName){
  rvector IN(6);
  triangle tempT;
  ifstream InFile(inFileName.c_str(), ios_base::in);
  while(InFile>>IN) {
    tempT.p1=complex(IN[1], IN[2]);
    tempT.p2=complex(IN[3], IN[4]);
    tempT.p3=complex(IN[5], IN[6]);
    tempT.wNum=-1;
    workList+=tempT;
  }
  InFile.close();
}

//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
int main(int argc, char * argv[])
{
  if ( argc != 5 ) {
    cout << "Syntax: " << argv[0] << " triangulation.txt sTol nTol mRoot" 
	 << endl << endl
	 << "triangulation.txt should contain a triangulation of the domain"
	 << endl 
	 << "sTol is the splitting tolerance, i.e the smallest diameter a "
	 << "box is split into."
	 << endl
	 << "10^{-nTol}, is the inflation box used for the result of Newtons "
	 << "method."
	 << endl
	 << "mRoot : check if clusters do not contain multiple roots, "
	 << "(slow if sTol large). 1=on, 0=off" 
	 << endl;
    exit(0);
  }
  cout << endl
       << "**********************************************************" << endl
       << "************* cZero - complex root finder** **************" << endl
       << endl << "Latest compilation: " << __DATE__ << ", " << __TIME__
       << endl << endl;

  string inFileName(argv[1]);
  real stol(atof(argv[2]));
  int ntolp(atoi(argv[3]));
  int mRoot(atoi(argv[4]));

  cout<< SetPrecision(17,16);

  List<triangle> workList;
  initiate(workList, inFileName);
  cout<<"|workList| = "<<Length(workList)<<endl;
  clock(START_TIMING);
  int totZero=0;
  
  ofstream OutFile("out.txt", ios_base::out);
  
  while(!IsEmpty(workList)) {
    triangle workDom=Pop(workList);
    bool fail(1);
    interval arg(0);
    try{
      //Only calculate the winding number if previously unknown
      if( workDom.wNum == -1 )
	arg=quad(workDom, fail, 1);
      else
	arg=interval(workDom.wNum);
    }
    catch(...){fail=0;}
  
    if(fail==0) {
      if (workDom.sizeOf()<stol) {
	cout<<"cannot integrate on : " <<endl; 
	workDom.display();
      }
      else
	splitAndStore(workDom, workList); 
    }
    else {
      if (ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) {
	workDom.wNum=abs((int)ceil(_double(Inf(arg))));
                 
	if (workDom.wNum==1) {
	  //Try to improve the enclosure
	  if(searchZero(workDom, 1, ntolp) || (workDom.sizeOf()<stol)) {
	    cout<<"There is: "<< workDom.wNum << " zero in the domain: " <<endl;
	    workDom.display();
	    totZero+=workDom.wNum;
	    OutFile<<Re(workDom.p1)<<" "<<Im(workDom.p1)<<" "<<
	      Re(workDom.p2)<<" "<<Im(workDom.p2)<<" "
		   <<Re(workDom.p3)<<" "<<Im(workDom.p3) <<" "<<workDom.wNum<<endl;
	  }
 	  else
 	    splitAndStore(workDom, workList);
	}
	else if(workDom.wNum>1) {
	  //Try to improve the enclosure in the case of multiple zeros
	  if ( (mRoot==1 && searchZero(workDom, workDom.wNum, ntolp) ) || (workDom.sizeOf()<stol) ) {
	    cout<<"There are: "<< workDom.wNum << " zeros in the domain: "<<endl;
	    workDom.display(); 
	    totZero+=workDom.wNum;
	    OutFile<<Re(workDom.p1)<<" "<<Im(workDom.p1)<<" "<<
	      Re(workDom.p2)<<" "<<Im(workDom.p2)<<" "
		   <<Re(workDom.p3)<<" "<<Im(workDom.p3) <<" "<<workDom.wNum<<endl;

	    if(mRoot==1) {
	      //try to disprove the existence of multiple roots
	      while(workDom.wNum>1 && !MultipleRoot(workDom, workDom.wNum)) {
		cout<<"There are no zeros of order "<< workDom.wNum <<" in the domain"<<endl;
		workDom.wNum--;
	      }
	    }
	  }
	  else
	    splitAndStore(workDom, workList);
	}
      }
    }
  }
  OutFile.close();
  cout<<"Total number of zeros = " << totZero << endl;
  clock(STOP_TIMING);
  return 0;
}
//--------------------------------------------------------------------------


