/*   File: cZero.cc
     compute Zeros of holomorphic functions
    
     The algorithm is based on a combination
     of validated winding number calculations
     and floating point Newton searches.

     Author: Tomas Johnson <johnson@math.uu.se>      
     Latest edit: 05/02/2008
     By: Tomas Johnson <johnson@math.uu.se>
*/

#include <fstream>
#include <cstdlib>
#include <iostream>
#include <string>
#include <stdexcept>
#include <time.h>
#include <unistd.h>
#include "list.h"

#include <cinterval.hpp>
#include <cimath.hpp>
#include "/home/nobackup/warwick/taylor/citaylor.hpp"

//#include "citaylor.hpp"

using namespace cxsc;
using namespace std;
using namespace taylor;

const interval PI(Pi());       // An enclosure of pi.
const cinterval I(interval(0), interval(1)); //An enclosure of i
const int RECNUM(10000);  //Maximal recursion depth
enum command {START_TIMING, SHOW_TIMING, STOP_TIMING};


////////////////////////////////////////////////////////////////
//Auxiliary power function, to be used for z^n, n\in N
//The reason is that there is no special method in citaylor
//to compute the power function for this case. I.e. z^n is calculated
//as exp(n*log(z)) which can cause (non-mathematical) branch cut 
//problems when the derivatives of z^n are calculated.
//This implementation is (perhaps) non-optimal, but gives a method to 
//enclose the derivatives of z^n without causing run-time errors.
//In case that the real part of z does not cut Re(z)=0
//AND the real part of the box is large enough (completely heuristic !)
//it computes the power function using the one included in citaylor.
//Otherwise the function computes z^n recursively as:
//           z^n = z^(mod(n,2))*sqr(z^(floor(n/2)));
///////////////////////////////////////////////////////////////////
citaylor aPow(const citaylor &z, const int &n) {
  if ( !(0.0 <= get_j_coef(z,0)) && (diam(Re(get_j_coef(z,0)))>0.00001) ) {
    citaylor res=power(z,n);
    return res;
  }
  
  if (n==1)
    return z;
  if (n==2)
    return sqr(z);
  int k=n/2;
  if (n==2*k) 
    return sqr(aPow(z,k));
  else
    return z*sqr(aPow(z,k));
  
}
//////////////////////////////////////////////////////////
//Parametrisation of the boundary for the higher order case
////////////////////////////////////////////////////////////////////////////

citaylor zoft(const int &id, const citaylor &t, const complex &llC, const complex &urC) {
  if(id==1){
    complex startP(Re(urC) , Im(llC));
    complex stopP(urC);
    return startP + t*(stopP-startP);
  }
  if(id==2) {
    complex startP(urC);
    complex stopP(Re(llC), Im(urC));
    return startP + t*(stopP-startP);
  }
  if(id==3) {
    complex startP(Re(llC),Im(urC));
    complex stopP(llC);
    return startP + t*(stopP-startP);
  }
  if(id==4) {
    complex startP(llC);
    complex stopP(Re(urC), Im(llC));
    return startP + t*(stopP-startP);
  }
  cout<<"ERROR"<<endl;
  exit(0);
}

////////////////////////////////////////////////////////////////////////////
//
// 'START_TIMING' starts the clock, 'STOP_TIMING' stops the clock,
// and prints the elapsed time in hh:mm:ss. 'SHOW_TIMING' prints the
// elapsed time.
//
void clock(const command &cmd)
{
  struct tm     *ptr;          // Used for timing the program
  static time_t  time_start;   // start time

  if ( cmd == START_TIMING ) {
    time_start = time(NULL);
    ptr = localtime(&time_start);
    cout << "Started computations: " << asctime(ptr);
  }
  else if ( cmd == SHOW_TIMING || cmd == STOP_TIMING ) {
    int weeks   = 0;
    int days    = 0;
    int hours   = 0;
    int minutes = 0;
    double seconds;
    time_t time_stop;

    time_stop = time(NULL);
    ptr = localtime(&time_stop);
    if ( cmd == STOP_TIMING )
      cout << "Completed computations: " << asctime(ptr);
    seconds = difftime(time_stop, time_start);
    cout << "Elapsed run time: ";
    // Convert seconds to vv:dd:hh:mm:ss
    while ( seconds >= 60 ) { seconds -= 60; minutes++; }
    while ( minutes >= 60 ) { minutes -= 60;   hours++; }
    while (   hours >= 24 ) {   hours -= 24;    days++; }
    while (    days >=  7 ) {    days -=  7;   weeks++; }
    if ( weeks > 0)
      cout << weeks << " week" << (weeks != 1 ? "s, " : ", ");
    if ( days > 0)
      cout << days << " day" << (days != 1 ? "s, " : ", ");
    if ( hours > 0)
      cout << hours << " hour" << (hours != 1 ? "s, " : ", ");
    if ( minutes > 0)
      cout << minutes << " minute" << (minutes != 1 ? "s, " : ", ");
    cout << seconds << " second" << (seconds != 1 ? "s." : ".") << endl;
  }
}

////////////////////////////////////////////////////////////////////////////
//The function for higher orders.
citaylor fcnF(const citaylor &z, bool &fOK) {
  //return aPow(z,11)-1;
  //  ENTER CONDITIONS ON f's domain!!
 //  if (0<=get_j_coef(z,0)) {
//     fOK=false;
//     return citaylor(get_order(z),1);
//   }
  //Herlocker, Ely:
  //return power(z,3) - sqr(z) - z - 2; //(1)
  //Some other examples
  //return power(z,3) - sqr(z) - z - 2+sin(10*sqr(z)); //(2)
  //return sin(z);  //(3)
  
  //Dellnitz, Schtze, Zheng
  // return power(z,50)+power(z,12)-5*sin(20*z)*cos(12*z)-1;  //(4)

  // return aPow(z,50)+aPow(z,12)-5*sin(20*z)*cos(12*z)-1;  //(4)
  //return MyPow(z,50)+MyPow(z,12)-5*sin(20*z)*cos(12*z)-1;  //(4)
  //return sqr(z)+interval(-19435)/interval(pow(10,6))*z+interval(100041)/interval(100)*
  //  exp(-interval(5)/interval(1000)*z)+interval(522463); //(5)

 //  interval A(166689);
//   A=A/interval(pow(10,7));
//   interval B(166659);
//   B=B/interval(pow(10,7));
//   interval C(-37485);
//   C=C/interval(pow(10,9));
//   interval D(166667);
//   D=D/interval(pow(10,7));
  // interval t(2.0);
//   return ( ((A+interval(212)/interval(pow(10,16))*z)*(D+interval(6.0)/interval(pow(10,16))*z))-
// 	   ((B+exp(-t*z)*(C+interval(6.0)/interval(pow(10,16))*z))*
// 	    (interval(1.0)/interval(60.0)+interval(6.0)/interval(pow(10,16))*exp(-t*z)*z)) ); //(6)

  return interval(70)*sqr(sqr(z)-2*z+3)*(z-1-interval(99)/interval(70)*I); //(7)
    //As in Ko et al
  //return sqr(sqr(z)+z+1)*aPow(z-1,4)*aPow(z*sqr(z)+sqr(z)+z+1,3)*(z-2)*aPow(z-4,4); //(18s)
  //return interval(1.0)-sqr(aPow(z,16)-1); //(11s)
  //return (sqr(z)-1)*(z-I);
}

//////////////////////////////////////////////////////////
//The function for order 1

cinterval FCNF(const cinterval &z) {
  //return sqr(sqr(z)+z+1)*power(z-1,4)*power(z*sqr(z)+sqr(z)+z+1,3)*(z-2)*power(z-4,4); //(18s)
  //return power(z,11)-1;
     return interval(70)*sqr(sqr(z)-2*z+3)*(z-1-interval(99)/interval(70)*I); //(7)

  // return power(z,50)+power(z,12)-5*sin(20*z)*cos(12*z)-1;  //(4)

  //return sqr(z)+interval(-19435)/interval(pow(10,6))*z+interval(100041)/interval(100)*
  // exp(-interval(5)/interval(1000)*z)+interval(522463); //(5)
  // return (sqr(z)-1)*(z-I);
}

///////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
//Compute the winding number, using 
//\sum_{discretisation}Arg(f(c_{i+1})/f(c_i))
interval simInt2(const int &id, const interval &t, const complex &llC, 
		 const complex &urC, bool &ok, const int &order) {
  cinterval rightleftF;
  if (order==1) {
    cinterval z, rightz, leftz;
    if(id==1){
      complex startP(Re(urC) , Im(llC));
      complex stopP(urC);
      z=startP + t*(stopP-startP);
      rightz=startP + Sup(t)*(stopP-startP);
      leftz=startP + Inf(t)*(stopP-startP);
    }
    if(id==2) {
      complex startP(urC);
      complex stopP(Re(llC), Im(urC));
      z=startP + t*(stopP-startP);
      rightz=startP + Sup(t)*(stopP-startP);
      leftz=startP + Inf(t)*(stopP-startP);
    }
    if(id==3) {
      complex startP(Re(llC),Im(urC));
      complex stopP(llC);
      z=startP + t*(stopP-startP);
      rightz=startP + Sup(t)*(stopP-startP);
      leftz=startP + Inf(t)*(stopP-startP);
    }
    if(id==4) {
      complex startP(llC);
      complex stopP(Re(urC), Im(llC));
      z=startP + t*(stopP-startP);
      rightz=startP + Sup(t)*(stopP-startP);
      leftz=startP + Inf(t)*(stopP-startP);
    } 
  
    if( 0.0<=Re(FCNF(z)) && 0.0<=Im(FCNF(z)) ){
      ok=false;
      return t;
    }
    
    rightleftF=FCNF(rightz)/FCNF(leftz);
  }
  //In the multiple case
  else {
    citaylor tA(order-1, t);
    citaylor tr(order-1, Sup(t));
    citaylor tl(order-1, Inf(t));
    citaylor z=zoft(id, tA, llC, urC);
    citaylor rightz=zoft(id, tr, llC, urC);
    citaylor leftz=zoft(id, tl, llC, urC);
    
    cinterval fVal=get_j_coef(fcnF(z,ok), order-1);
    if( 0.0<=Re(fVal) && 0.0<=Im(fVal) ){
      ok=false;
      return t;
    }
    rightleftF=get_j_coef(fcnF(rightz,ok), order-1)/get_j_coef(fcnF(leftz, ok),order-1);

  }

  //Forces initial, minimum discretisation
  //Arg, is the principal value argument function
  if (Inf(Re(rightleftF))<=0.0 && 0<=Im(rightleftF) ) {
    ok=false;
    return t;
  }
  
  return Arg(rightleftF)/(2*PI);
}

//////////////////////////////////////////////////////////////////////////////////////////////
//fail =-1, possible zero on the contour
//fail = 0, to many iterations
//fail = 1, OK
//Run each side recursively until the enclosure is sufficiently good

interval quadSide(const int &id, const interval &t, const complex &llC, const complex &urC, 
		  const real &tol, int &fail, const int &rec, const int &order) {
  if (fail==-1)
    return t;
  
  if (fail==0 || (rec>RECNUM) ) {
    fail=0;
    return t;
  } 
    
  bool ok(true);
  interval res=simInt2(id, t, llC, urC, ok, order);

  if (!ok && rec>=RECNUM) {
    fail=-1;
    return t;
  }
  
  if(!ok || diam(res)>tol) {
    res=quadSide(id, interval(Inf(t),mid(t)), llC, urC, tol/2, fail, rec+1, order)
      +quadSide(id, interval(mid(t), Sup(t)), llC, urC, tol/2, fail, rec+1, order); 
  }
 
  return res;
}


///////////////////////////////////////////////////
//Compute the contribution to the winding number
//from one side at the time
///////////////////////////////////////////////

interval quad(const cinterval &Dom, int &fail, const int &order) {
  interval res(0);
  complex urC(Sup(Re(Dom)), Sup(Im(Dom)));
  complex llC(Inf(Re(Dom)), Inf(Im(Dom)));
  real tol(0.25);
  for(int id=1; id<=4; id++) 
    res+=quadSide(id, interval(0,1), llC, urC, tol, fail, 0, order);
  
  return res;
}

////////////////////////////////////////////////////////////////////////////
//
// Bisects in both components, and stores all four squares.
//
void splitAndStore(const cinterval &z, List<cinterval> &workList)
{
  interval reZ = Re(z);
  interval imZ = Im(z);

  workList+= cinterval(interval(Inf(reZ), mid(reZ)), interval(Inf(imZ), mid(imZ)));
  workList+= cinterval(interval(Inf(reZ), mid(reZ)), interval(mid(imZ), Sup(imZ)));
  workList+= cinterval(interval(mid(reZ), Sup(reZ)), interval(Inf(imZ), mid(imZ)));
  workList+= cinterval(interval(mid(reZ), Sup(reZ)), interval(mid(imZ), Sup(imZ)));
}


////////////////////////////////////////////////////////////////////////////
//Floating point Newton method, increases the speed of the algorithm
/////////////////////////////////////////////////////////////
bool searchZero(cinterval &Dom, const int &NZ, const int &ntolp) {
  if (NZ>100)  //Don't search for a zero of too high order
    return false;
  try {
    real eps(pow(10,-ntolp));
    citaylor z(NZ, mid(Dom));
    bool fOK(true);
    citaylor f=fcnF(z,fOK);
    if (!fOK)
      return false;
    
    complex zz;
    if (0<=get_j_coef(f,NZ))
      return false;
    else
      zz=mid(Dom)-mid(get_j_derive(f,NZ-1))/mid(get_j_derive(f,NZ));
    if (!(zz<=Dom))
      return false;
    
    int i=0;
    int nits=max(5,ntolp);
    while(i<nits && zz<=Dom) {
      citaylor zt(NZ,zz);
      f=fcnF(zt, fOK);
      if (!fOK || 0<=get_j_coef(f,NZ) ){
	i=nits;
      }
      else {
	complex zztemp=zz-mid(get_j_derive(f,NZ-1))/mid(get_j_derive(f,NZ));
	if(zztemp<=Dom) {
	  zz=zztemp;
	  i=i+1;
	}
	else{
	  i=nits;
	}
      }
    }
    
    int fail(1);
    cinterval Dom2;
    interval arg;
    Dom2=(zz+cinterval(interval(-eps,eps),interval(-eps,eps)))&Dom;
    //filib doesn't throw ! for example 4.
    if(0<=Im(Dom2)) {
      return false;
    }
    /////
    arg=quad(Dom2, fail, 1);
    if(fail!=1)
      return false;
    
    if ((ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) && (floor(_double(Sup(arg)))==NZ) ) {
      Dom=Dom2;
      return true;
    }
  }
  catch(...){}
  return false;
}
////////////////////////////////////////////////////////////////////////////
//Computes the winding number for derivative number (NZ-1)
bool MultipleRoot(const cinterval &Dom, const int &NZ) {
  if(NZ>10)  //Do not compute multiple roots of too high order
    return 1;
  int fail(1);

  interval arg=quad(Dom, fail, NZ);
  if(fail!=1)
    return 1;
  if (ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) {
    int resNZ=(int)ceil(_double(Inf(arg)));

    if(resNZ!=0)
      return 1;
    else
      return 0;
  }
  else
    return 1;
  
}

/////////////////////////////////////////////////////////////////////
int main(int argc, char * argv[])
{
  if ( argc != 8 ) {
    cout << "Syntax: " << argv[0] << " inf(real) sup(real) inf(im) sup(im) sTol nTol mRoot" << endl<<endl
	 << "Search box is: [inf(real), sup(real)]+i[inf(im), sup(im)]"<<endl
	 << "sTol is the splitting tolerance, i.e the smallest diameter a box is split into."<<endl
	 << "10^{-nTol}, is the inflation box used for the result of Newtons method."<<endl
	 << "mRoot : check if clusters do not contain multiple roots, (slow if sTol large). 1=on, 0=off"<<endl;
    exit(0);
  }
  cout << endl
       << "**********************************************************" << endl
       << "************* cZero - complex root finder** **************" << endl
       << endl << "Latest compilation: " << __DATE__ << ", " << __TIME__
       << endl << endl;

  interval realDom(atof(argv[1]),atof(argv[2]) );
  interval imDom(atof(argv[3]),atof(argv[4]) );
  cinterval Dom(realDom, imDom);
  real stol(atof(argv[5]));
  int ntolp(atoi(argv[6]));
  int mRoot(atoi(argv[7]));
  cout<< SetPrecision(17,16);
  List<cinterval> workList, tempList;
 
  clock(START_TIMING);

  
  workList+=Dom;
  while(!IsEmpty(workList)) {
    cinterval workDom=Pop(workList);
    int fail(1);
    interval arg(0);
    try{
      arg=quad(workDom, fail, 1);
    }
    catch(...){fail=0;}

    if(fail==-1){
      cout<<"zero on contour, inflating!"<<endl;
      workList+=cinterval(interval(Inf(Re(workDom))-stol/2, Sup(Re(workDom))+stol/2) , 
			 interval(Inf(Im(workDom))-stol/2, Sup(Im(workDom))+stol/2) );
    }
    else if(fail==0) {
      if (diam(abs(workDom))<stol)
	cout<<"cannot integrate on : " <<Re(workDom)<<" + i"<<Im(workDom)<<endl;
      else
	splitAndStore(workDom, workList); 
    }
    else {
      if (ceil(_double(Inf(arg)))==floor(_double(Sup(arg)))) {
	int NZ=(int)ceil(_double(Inf(arg)));
	
	if(workDom==Dom)
	  cout<<"Total number of zeros = "<<NZ<<endl;
	if (NZ==1) {
	  if(searchZero(workDom,1, ntolp) || (diam(abs(workDom))<stol))
	    cout<<"There is: "<< NZ << " zero in the domain " <<Re(workDom)<<" + i"<<Im(workDom)<<endl;
	  else
	    splitAndStore(workDom, workList);
	}
	else if(NZ>1) {
	  if ( (mRoot==1 && searchZero(workDom,NZ , ntolp)) || diam(abs(workDom))<stol) {
	    cout<<"There are: "<< NZ << " zeros in the domain " <<Re(workDom)<<" + i"<<Im(workDom)<<endl; 
	    if(mRoot==1) {
	      while(NZ>1 && !MultipleRoot(workDom, NZ)) {
		cout<<"There are no zeros of order "<< NZ <<" in the domain"<<endl;
		NZ--;
	      }
	    }
	  }
	  else {
	    splitAndStore(workDom, workList);}
	}
      }
    }
  }
  clock(STOP_TIMING);
  return 0;
}

////////////////////////////////////////////////////////////////////////////

